# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""A lightweight Python library for anonymizing tabular data."""

from aindo.anonymize import version as aindo_version
from aindo.anonymize.config import Config
from aindo.anonymize.pipeline import AnonymizationPipeline

__version__ = aindo_version.__version__

__all__ = ["__version__", "Config", "AnonymizationPipeline"]



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of the binning technique."""

from typing import ClassVar, Sequence

import pandas as pd

from aindo.anonymize.techniques.base import BaseSingleColumnTechnique


class Binning(BaseSingleColumnTechnique):
    """Implements binning for numerical columns.

    Binning works by grouping numerical values into discrete bins,
    allowing for data generalization by replacing individual values with their corresponding bin ranges.

    Attributes:
        bins: The bin edges or number of bins to use.

    Examples:
        An integer bins will form equal-width bins.
        >>> ages = pd.Series([10, 15, 13, 12, 23, 25, 28, 59, 60])
        >>> Binning(bins=3).anonymize_column(ages)
        [(9.95, 26.667], (9.95, 26.667], (9.95, 26.667], ...
        Categories (3, interval[float64, right]): [(9.95, 26.667] < (26.667, 43.333] < (43.333, 60.0]]

        A list of ordered bin edges will assign an interval for each variable.
        >>> ages = pd.Series([10, 15, 13, 12, 23, 25, 28, 59, 60])
        >>> Binning(bins=[0, 18, 35, 70]).anonymize_column(ages)
        [(0, 18], (0, 18], (0, 18], (0, 18], (18, 35], ...
        Categories (3, interval[int64, right]): [(0, 18] < (18, 35] < (35, 70]]
    """

    preserve_type: ClassVar[bool] = False

    bins: int | Sequence[int] | Sequence[float]

    def __init__(self, bins: int | Sequence[int] | Sequence[float]) -> None:
        super().__init__()
        self.bins = bins

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        return pd.cut(col, self.bins, labels=None)



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of the top/bottom coding technique."""

from typing import Any, ClassVar

import numpy as np
import pandas as pd

from aindo.anonymize.techniques.base import BaseSingleColumnTechnique


class TopBottomCodingNumerical(BaseSingleColumnTechnique):
    """Implements top/bottom coding for numerical columns.

    This technique caps values above the `(1 - q/2)` quantile (top coding) and raises
    values below the `(q/2)` quantile (bottom coding). The threshold parameter `q` specifies the
    total proportion of extreme values to code (e.g., `q=0.1` applies top/bottom coding to 5% each).

    Either the threshold `q` or both quantile values (`lower_value` and `upper_value`) must be provided,
    but not both. If `lower_value` and `upper_value` are used, they must be specified together.

    Attributes:
        q: Proportion controlling the extent of top/bottom coding, between 0 and 1.
        lower_value: Input data quantile value at q/2.
        upper_value: Input data quantile value at (1- q/2).
    """

    q: float | None = None
    lower_value: float | None = None
    upper_value: float | None = None

    def __init__(
        self, q: float | None = None, lower_value: float | None = None, upper_value: float | None = None
    ) -> None:
        super().__init__()
        if (q is not None) and (not 0 <= q <= 1):
            raise ValueError("The threshold 'q' should be a value between 0 and 1")
        if (lower_value is None) != (upper_value is None):
            raise ValueError(
                "Quantile values (lower_value and upper_value) must be provided together or omitted entirely"
            )
        if (q is None) == (lower_value is None):
            raise ValueError("Either the threshold 'q' or the quantile values must be provided, but not both")
        self.q = q
        self.lower_value = lower_value
        self.upper_value = upper_value

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        lower_value: float
        upper_value: float
        if (self.lower_value is not None) and (self.upper_value is not None):
            lower_value = self.lower_value
            upper_value = self.upper_value
        else:
            assert self.q is not None
            lower_value = (
                col.quantile(self.q / 2, interpolation="nearest") if self.lower_value is None else self.lower_value
            )
            upper_value = (
                col.quantile(1 - self.q / 2, interpolation="nearest") if self.upper_value is None else self.upper_value
            )

        return col.clip(lower=lower_value, upper=upper_value)


class TopBottomCodingCategorical(BaseSingleColumnTechnique):
    """Implements top/bottom coding for categorical columns.

    Categories representing less or equal than `q` of the total data are replaced with the `other_label`
    (e.g.: q=0.01 represents the 1%).

    Attributes:
        q: A proportion controlling the extent of top/bottom coding, between 0 and 1.
        other_label: The new category to replace rare categories with. Default is "OTHER".
        rare_categories: A list of rare categories to be replaced.
            This can be used instead of the `q` parameter to explicitly specify
            which categories should be replaced with `other_label`.
    """

    preserve_type: ClassVar[bool] = False

    q: float | None = None
    other_label: Any = "OTHER"
    rare_categories: list[Any] | None = None

    def __init__(
        self, q: float | None = None, other_label: Any = "OTHER", rare_categories: list[Any] | None = None
    ) -> None:
        super().__init__()
        if (q is not None) and (not 0 <= q <= 1):
            raise ValueError("The value 'q' should be a value between 0 and 1.")
        if other_label is None:
            raise ValueError("The new category cannot be None")
        if (q is None) == (rare_categories is None):
            raise ValueError("Either 'q' or 'rare_categories' must be provided, but not both")
        self.q = q
        self.other_label = other_label
        self.rare_categories = rare_categories

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        out: pd.Series
        if col.dtype.name == "category":
            out = col.copy()
        else:
            out = col.astype("category")

        rare_categories: list[Any]
        if self.rare_categories is None:
            value_counts: pd.Series = col.value_counts(normalize=True)
            self.frequencies: dict = dict(zip(value_counts.index.to_list(), value_counts.values))
            rare_categories = value_counts[value_counts <= self.q].index.to_list()
        else:
            rare_categories = self.rare_categories

        if not rare_categories:
            return out

        out = out.cat.remove_categories(rare_categories)
        out = out.cat.add_categories(self.other_label)
        out.fillna(self.other_label, inplace=True)
        out[col.isna()] = np.nan

        return out



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of the swapping technique."""

import pandas as pd

from aindo.anonymize.techniques.base import BaseSingleColumnTechnique
from aindo.anonymize.techniques.common import AlphaProbability, Seeder, SeedT


class Swapping(BaseSingleColumnTechnique, Seeder, AlphaProbability):
    """Implements swapping.

    Swapping rearranges data by shuffling values,
    ensuring that individual values remain present but are generally not in their original position.
    The process is controlled by the `alpha` parameter,
    representing the probability of a row being swapped with another.

    Attributes:
        alpha: The perturbation intensity, a value in the range [0, 1].
    """

    def __init__(self, alpha: float, **kwargs: SeedT) -> None:
        Seeder.__init__(self, **kwargs)
        AlphaProbability.__init__(self, alpha)

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        if self.alpha == 0:
            return col.copy()

        rows_to_swap = [idx for idx in col.index if self.generator.random() <= self.alpha]
        out: pd.Series = col.copy()

        for idx in rows_to_swap:
            new_idx: int = self.generator.choice(rows_to_swap[:idx] + rows_to_swap[idx + 1 :], 1)[0]

            out[idx], out[new_idx] = out[new_idx], out[idx]

        return out



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of the mocking technique."""

from functools import partial
from typing import Any, Callable, ClassVar, Literal

import pandas as pd
from faker import Faker
from faker.typing import SeedType

from aindo.anonymize.techniques.base import BaseSingleColumnTechnique

# fmt: off
MockingGeneratorMethods = Literal[
    "aba","address","administrative_unit","am_pm","android_platform_token","ascii_company_email","ascii_email",
    "ascii_free_email","ascii_safe_email","bank_country","bban","binary","boolean","bothify","bs","building_number",
    "catch_phrase","century","chrome","city","city_prefix","city_suffix","color","color_name","company",
    "company_email", "company_suffix","coordinate","country","country_calling_code","country_code",
    "credit_card_expire", "credit_card_full","credit_card_number","credit_card_provider",
    "credit_card_security_code","cryptocurrency", "cryptocurrency_code","cryptocurrency_name","csv","currency",
    "currency_code","currency_name","currency_symbol","current_country","current_country_code","date","date_between",
    "date_between_dates","date_object","date_of_birth","date_this_century","date_this_decade","date_this_month",
    "date_this_year","date_time","date_time_ad","date_time_between","date_time_between_dates",
    "date_time_this_century","date_time_this_decade","date_time_this_month","date_time_this_year","day_of_month",
    "day_of_week","dga","domain_name","domain_word","dsv","ean","ean13","ean8","ein","email","emoji","enum",
    "file_extension","file_name","file_path","firefox","first_name","first_name_female","first_name_male",
    "first_name_nonbinary","fixed_width","free_email","free_email_domain","future_date","future_datetime",
    "hex_color","hexify","hostname","http_method","iana_id","iban","image","image_url","internet_explorer",
    "invalid_ssn","ios_platform_token","ipv4","ipv4_network_class","ipv4_private","ipv4_public","ipv6","isbn10",
    "isbn13","iso8601","itin","job","json","json_bytes","language_code","language_name","last_name",
    "last_name_female","last_name_male","last_name_nonbinary","latitude","latlng","lexify","license_plate",
    "linux_platform_token","linux_processor","local_latlng","locale","localized_ean","localized_ean13",
    "localized_ean8","location_on_land","longitude","mac_address","mac_platform_token","mac_processor",
    "md5","military_apo","military_dpo","military_ship","military_state","mime_type","month","month_name",
    "msisdn","name","name_female","name_male","name_nonbinary","nic_handle","nic_handles","null_boolean",
    "numerify","opera","paragraph","paragraphs","password","past_date","past_datetime","phone_number",
    "port_number","postalcode","postalcode_in_state","postalcode_plus4","postcode","postcode_in_state",
    "prefix","prefix_female","prefix_male","prefix_nonbinary","pricetag","profile","providers","psv",
    "pybool","pydecimal","pydict","pyfloat","pyint","pyiterable","pylist","pyset","pystr","pystr_format",
    "pystruct","pytimezone","pytuple","random_choices","random_digit","random_digit_not_null",
    "random_digit_not_null_or_empty","random_digit_or_empty","random_element","random_elements","random_int",
    "random_letter","random_letters","random_lowercase_letter","random_number","random_sample",
    "random_uppercase_letter","randomize_nb_elements","rgb_color","rgb_css_color","ripe_id","safari",
    "safe_color_name","safe_domain_name","safe_email","safe_hex_color","secondary_address","sentence",
    "sentences","sha1","sha256","simple_profile","slug","ssn","state","state_abbr","street_address",
    "street_name","street_suffix","suffix","suffix_female","suffix_male","suffix_nonbinary","swift",
    "swift11","swift8","tar","text","texts","time","time_delta","time_object","time_series","timezone","tld","tsv",
    "unix_device","unix_partition","unix_time","upc_a","upc_e","uri","uri_extension","uri_page","uri_path","url",
    "user_agent","user_name","uuid4","windows_platform_token","word","words","year","zip","zipcode",
    "zipcode_in_state","zipcode_plus4",
]
"""List all available generator methods from Faker (`fake`)."""
# fmt: on


class Mocking(BaseSingleColumnTechnique):
    """Implements mocking.

    Mocking generates realistic mock data for various fields such as names, addresses, emails, and more.
    It leverages the `faker` library to produce customizable, locale-aware fake data.

    Missing values (`np.NaN`, `None`, `pd.NA`, `pd.NaT`) are replaced.

    Attributes:
        data_generator: Faker's generator method ("fake") used to generate data (e.g., name, email).
        seed: A seed to initialize numpy `Generator`.
    """

    preserve_type: ClassVar[bool] = False

    data_generator: "MockingGeneratorMethods"
    seed: SeedType = None

    def __init__(
        self,
        data_generator: "MockingGeneratorMethods",
        seed: SeedType = None,
        faker_kwargs: dict[str, Any] | None = None,
        faker_generator_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Initializes the Mocking technique.

        Args:
            data_generator: Faker's generator method ("fake") used to generate data (e.g., name, email).
            seed: A seed to initialize numpy `Generator`.
            faker_kwargs: Additional arguments passed to the main Faker object (proxy class).
            faker_generator_kwargs: Additional arguments passed to the Faker's generator method.
        """
        super().__init__()
        fake = Faker(**(faker_kwargs or {}))
        fake_method = getattr(fake, data_generator, None)
        if fake_method is None:
            raise ValueError(f"Faker's generator method '{data_generator}' not found")

        self.data_generator = data_generator
        self._fake: Faker = fake
        self.seed = seed
        if seed is not None:
            self._fake.seed_instance(self.seed)

        self._fake_method: Callable[[], Any] = partial(fake_method, **(faker_generator_kwargs or {}))

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        return pd.Series([self._fake_method() for _ in col])



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of the perturbation technique."""

from __future__ import annotations

from abc import ABC
from typing import Generic, Literal, TypeVar, cast

import numpy as np
import pandas as pd
from sklearn.preprocessing import QuantileTransformer

from aindo.anonymize.techniques.base import BaseSingleColumnTechnique
from aindo.anonymize.techniques.common import AlphaProbability, Seeder, SeedT

SamplingMode = Literal["uniform", "weighted"]
NumericsT = TypeVar("NumericsT", int, float)


class BasePerturbation(BaseSingleColumnTechnique, Seeder, AlphaProbability, ABC):
    """Base class for perturbation implementations."""

    def __init__(self, alpha: float, **kwargs: SeedT):
        Seeder.__init__(self, **kwargs)
        AlphaProbability.__init__(self, alpha)


class PerturbationNumerical(BasePerturbation, Generic[NumericsT]):
    """Implements perturbation for numerical columns.

    Perturbation consists of modify each value based on the specified perturbation intensity (alpha)
    and replacement strategy.
    It supports two modes of replacement: uniform sampling and distribution-preserving sampling.

    Attributes:
        alpha: The perturbation intensity, a value in the range [0, 1].
            - `alpha=0`: No perturbation; values remain unchanged.
            - `alpha=1`: Maximum perturbation; values are fully replaced according to the specified sampling mode.
        sampling_mode: The strategy used to sample replacement values:
            - `uniform`: Values are perturbed with random values uniformly sampled from the range [min, max].
            - `weighted`: Values are perturbed in a way to keep the original distribution.
        perturbation_range: A tuple[min, max] within which random values are sampled.
            If not set, the range is automatically computed as the minimum and maximum of the input data.
    """

    sampling_mode: SamplingMode = "weighted"
    perturbation_range: tuple[NumericsT, NumericsT] | None = None

    def __init__(
        self,
        alpha: float,
        sampling_mode: SamplingMode = "weighted",
        perturbation_range: tuple[NumericsT, NumericsT] | None = None,
        **kwargs: SeedT,
    ) -> None:
        super().__init__(alpha, **kwargs)
        self.sampling_mode = sampling_mode
        self.perturbation_range = perturbation_range

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        if self.alpha == 0.0:
            return col.copy()

        if self.sampling_mode == "weighted":
            random_state: int | None = (
                self.seed
                if self.seed is None or isinstance(self.seed, int)
                else cast(int, self.generator.integers(0, 2**32, dtype=np.uint32))
            )
            transformer = QuantileTransformer(
                output_distribution="normal", n_quantiles=min(1_000, col.shape[0]), random_state=random_state
            )
            col_r: np.ndarray = transformer.fit_transform(col.to_numpy().reshape(-1, 1))
            random_values: np.ndarray = self.generator.normal(size=col_r.shape)
            col_r = np.sqrt(1 - self.alpha) * col_r + np.sqrt(self.alpha) * random_values
            col_r = cast(np.ndarray, transformer.inverse_transform(col_r)).reshape(-1)

            if np.issubdtype(col.dtype.type, np.integer):
                col_r = col_r.round()
            return pd.Series(col_r, dtype=col.dtype)

        else:
            perturbation_range: tuple[NumericsT, NumericsT]
            if self.perturbation_range is None:
                perturbation_range = (col.min(), col.max())
            else:
                perturbation_range = self.perturbation_range
            min_, max_ = perturbation_range

            random_values: np.ndarray = self.generator.uniform(low=min_, high=max_, size=col.size)
            col_out: pd.Series = (1 - self.alpha) * col + self.alpha * random_values
            return col_out


class PerturbationCategorical(BasePerturbation):
    """Implements perturbation for categorical columns.

    Perturbation consists of replacing values with randomized alternatives based on the specified
    sampling mode and perturbation intensity (alpha).
    It supports two modes of replacement: uniform sampling and distribution-preserving sampling.

    Attributes:
        alpha: The perturbation intensity, a value in the range [0, 1].
            - `alpha=0`: No perturbation; values remain unchanged.
            - `alpha=1`: Maximum perturbation; values are fully replaced according to the specified sampling mode.
        sampling_mode: The strategy used to sample replacement values:
            - `uniform`: Replaces values with others chosen uniformly at random.
            - `weighted`: Replaces values based on their original distribution.
        frequencies: Optional mapping of unique values to their relative frequencies, used for weighted sampling mode.
            Automatically computed if not provided.
    """

    sampling_mode: SamplingMode = "weighted"
    frequencies: dict[str, float] | None = None

    def __init__(
        self,
        alpha: float,
        sampling_mode: SamplingMode = "weighted",
        frequencies: dict[str, float] | None = None,
        **kwargs: SeedT,
    ) -> None:
        super().__init__(alpha, **kwargs)
        self.sampling_mode = sampling_mode
        self.frequencies = frequencies

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        frequencies: dict[str, float]
        if self.frequencies is None:
            value_counts: pd.Series = col.value_counts(normalize=True)
            frequencies = dict(zip(value_counts.index.to_numpy(), value_counts.values))
        else:
            frequencies = self.frequencies

        categories: np.ndarray = np.array(list(frequencies))
        probabilities: np.ndarray | None = (
            np.array(list(frequencies.values())) if self.sampling_mode == "weighted" else None
        )
        rand_values = self.generator.choice(categories, size=col.size, p=probabilities)
        mask: np.ndarray = col.notna().to_numpy() & (self.generator.random(col.size) <= self.alpha)

        return pd.Series(np.where(mask, rand_values, col), dtype="category")



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of the data nulling technique."""

from typing import Any, ClassVar

import pandas as pd

from aindo.anonymize.techniques.base import BaseSingleColumnTechnique


class DataNulling(BaseSingleColumnTechnique):
    """Implements data nulling.

    Data nulling replaces the original data with a constant value.
    Missing values (`np.NaN`, `None`, `pd.NA`, `pd.NaT`) are also replaced.

    Attributes:
        constant_value: The value that will replace the original data. Default to None.
    """

    preserve_type: ClassVar[bool] = False

    constant_value: Any = None

    def __init__(self, constant_value: Any = None) -> None:
        super().__init__()
        self.constant_value = constant_value

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        return pd.Series([self.constant_value for _ in col])



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of the identity technique."""

import pandas as pd

from aindo.anonymize.techniques.base import BaseTechnique


class Identity(BaseTechnique):
    """Identity technique.

    Leaves the original data untouched.
    This special technique is particularly useful in a declarative approach (see documentation).
    """

    def __init__(self) -> None:
        super().__init__()

    def anonymize(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """Anonymize the input data using the identity technique.

        Args:
            dataframe: The input data to be anonymized.

        Returns:
            The anonymized version of the input data.
        """
        return dataframe.copy()



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of all anonymization techniques."""

from aindo.anonymize.techniques.binning import Binning
from aindo.anonymize.techniques.char_masking import CharacterMasking
from aindo.anonymize.techniques.data_nulling import DataNulling
from aindo.anonymize.techniques.hashing import KeyHashing
from aindo.anonymize.techniques.identity import Identity
from aindo.anonymize.techniques.mocking import Mocking, MockingGeneratorMethods
from aindo.anonymize.techniques.perturbation import PerturbationCategorical, PerturbationNumerical
from aindo.anonymize.techniques.swapping import Swapping
from aindo.anonymize.techniques.top_bottom_coding import TopBottomCodingCategorical, TopBottomCodingNumerical

__all__ = [
    "Binning",
    "CharacterMasking",
    "DataNulling",
    "KeyHashing",
    "Identity",
    "Mocking",
    "MockingGeneratorMethods",
    "PerturbationCategorical",
    "PerturbationNumerical",
    "TopBottomCodingCategorical",
    "Swapping",
    "TopBottomCodingNumerical",
]



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of the hashing technique."""

import hashlib
import hmac
import secrets
from base64 import b64encode
from typing import Any, ClassVar

import pandas as pd

from aindo.anonymize.techniques.base import BaseSingleColumnTechnique


class KeyHashing(BaseSingleColumnTechnique):
    """Implements key-based hashing.

    Data values are hashed using HMAC with a cryptographic key and the chosen hashing algorithm
    (defaults to SHA-256). The resulting hash is then encoded using Base64.
    The de-identified values have always a uniform length.

    Attributes:
        key: The cryptographic key used for hashing.
        salt: An optional salt that can be added to the value before hashing. Defaults to None.
        hash_name: The hashing algorithm to use, compatible with `hashlib.new()`. Defaults to "sha256".
    """

    preserve_type: ClassVar[bool] = False

    key: str
    salt: str | None = None
    hash_name: str = "sha256"

    def __init__(self, key: str, salt: str | None = None, hash_name: str = "sha256") -> None:
        super().__init__()
        self.key = key
        self.salt = salt
        self.hash_name = hash_name

        if hash_name not in hashlib.algorithms_available:
            raise ValueError(f"Algorithm '{hash_name}' not available")

        _salt: str = salt or ""
        self._base_hmac_obj: hmac.HMAC = hmac.new(key.encode(), _salt.encode(), hashlib.sha256)

    def _compute_hash(self, value: Any) -> str:
        hmac_obj = self._base_hmac_obj.copy()
        hmac_obj.update(str(value).encode() if not isinstance(value, bytes) else value)
        return b64encode(hmac_obj.digest()).decode()

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        return pd.Series([self._compute_hash(value) if pd.notna(value) else value for value in col], dtype="str")

    @classmethod
    def generate_salt(cls) -> str:
        """Generates a random salt.

        Returns:
            str: A random salt.
        """
        return secrets.token_hex(16)



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""A high-level interface for orchestrating the anonymization process."""

import itertools
from typing import cast

import pandas as pd

from aindo.anonymize.config import BaseSpec, Config


class AnonymizationPipeline:
    """A high-level interface for orchestrating the anonymization process.

    This class provides a quick way to apply anonymization techniques to a dataset,
    allowing users to run anonymization pipelines with minimal setup.

    The resulting dataset will contain only the columns where a technique has been applied.
    To retain a column with its original data, use the Identity technique.

    Attributes:
        config: Configuration that specifies the anonymization steps to execute.
    """

    config: Config

    def __init__(self, config: Config) -> None:
        self.config = config

    def run(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """Runs the anonymization steps defined in the configuration against the input data.

        Args:
            dataframe: The input data to be anonymized.

        Returns:
            The anonymized version of the input data.
        """
        if len(self.config.steps) == 0:
            return pd.DataFrame()

        with pd.option_context("mode.copy_on_write", True):
            used_cols: set[str] = set(
                itertools.chain.from_iterable(step.columns or dataframe.columns for step in self.config.steps)
            )
            unused_cols: set[str] = set(dataframe.columns.to_list()) - used_cols
            result: pd.DataFrame = dataframe.drop(labels=cast(list, unused_cols), axis=1)

            for step in self.config.steps:
                method: BaseSpec = step.method
                if step.columns is None:
                    result = method.anonymize(result)
                    continue

                anonymized: pd.DataFrame = method.anonymize(result.loc[:, step.columns])
                if method.preserve_type:
                    result.loc[:, step.columns] = anonymized
                else:
                    for col_name in step.columns:
                        result[col_name] = anonymized[col_name]

            return result



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Mixin class implementations."""

from numpy.random import Generator, default_rng

SeedT = int | Generator | None


class Seeder:
    """Mixin class which provides easy access to numpy `Generator`.

    Attributes:
        seed: A seed to initialize numpy `Generator`.
    """

    seed: SeedT = None

    def __init__(self, seed: SeedT = None) -> None:
        self.seed = seed
        self._generator: Generator = default_rng(seed)

    @property
    def generator(self) -> Generator:
        """Seeded numpy `Generator`."""
        return self._generator


class AlphaProbability:
    """Mixin class to assist techniques that depend on an alpha value.

    Attributes:
        alpha: A float value in the range [0, 1].
    """

    alpha: float

    def __init__(self, alpha: float) -> None:
        if not 0 <= alpha <= 1:
            raise ValueError("alpha must be between 0 and 1.")
        self.alpha = alpha



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Configuration for the high-level interface `aindo.anonymize.AnonymizationPipeline`."""

import re
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Union, get_origin, get_type_hints

from aindo.anonymize.techniques import (
    Binning,
    CharacterMasking,
    DataNulling,
    Identity,
    KeyHashing,
    Mocking,
    PerturbationCategorical,
    PerturbationNumerical,
    Swapping,
    TopBottomCodingCategorical,
    TopBottomCodingNumerical,
)
from aindo.anonymize.techniques.base import BaseTechnique

ALL_TECHNIQUES: list[type[BaseTechnique]] = [
    Binning,
    CharacterMasking,
    DataNulling,
    KeyHashing,
    Identity,
    Mocking,
    PerturbationCategorical,
    PerturbationNumerical,
    Swapping,
    TopBottomCodingCategorical,
    TopBottomCodingNumerical,
]


class TechniqueType(str, Enum):
    """An enumeration of anonymization techniques."""

    BINNING = "binning"
    CHARACTER_MASKING = "character_masking"
    DATA_NULLING = "data_nulling"
    KEY_HASHING = "key_hashing"
    IDENTITY = "identity"
    MOCKING = "mocking"
    PERTURBATION_CATEGORICAL = "perturbation_categorical"
    PERTURBATION_NUMERICAL = "perturbation_numerical"
    SWAPPING = "swapping"
    TOP_BOTTOM_CODING_CATEGORICAL = "top_bottom_coding_categorical"
    TOP_BOTTOM_CODING_NUMERICAL = "top_bottom_coding_numerical"


def _get_type_from_class(cls: type[BaseTechnique]) -> TechniqueType:
    """Get the `TechniqueType` for the given technique or spec class."""
    type_name: str = re.sub(r"([a-z])([A-Z])", r"\1_\2", cls.__name__)
    type_name = type_name.upper()
    if type_name.endswith("_SPEC"):
        type_name = type_name.removesuffix("_SPEC")
    return TechniqueType[type_name]


class BaseSpec(BaseTechnique):
    """Base class for "Spec" classes.

    Attributes:
        type: Specifies which technique this configuration applies to.
    """

    _spec_fields: ClassVar[tuple[str, ...]]
    type: TechniqueType

    def __eq__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        return all(getattr(self, attr) == getattr(other, attr) for attr in self._spec_fields)


def _process_technique_class(cls: type[BaseTechnique]) -> type[BaseSpec]:
    """Derive a new class from a technique class by adding a type attribute.

    For example, a class derived from `DataNulling` will be equivalent to:
    ```python
    class DataNullingSpec(DataNulling):
        type: Literal[TechniqueType.DATA_NULLING] = TechniqueType.DATA_NULLING

    ```
    """
    _type: TechniqueType = _get_type_from_class(cls)

    field_annotations: dict[str, Any] = get_type_hints(cls)
    field_annotations["type"] = Literal[_type]

    field_values: dict[str, Any] = {"type": _type}
    field_values["_spec_fields"] = tuple(
        [
            name
            for name, _type in field_annotations.items()
            if not (name.startswith("_") or get_origin(_type) is ClassVar)
        ]
    )

    return type(
        f"{cls.__name__}Spec",
        (
            BaseSpec,
            cls,
        ),
        {**field_values, "__annotations__": field_annotations},
    )


ALL_TECHNIQUES_SPEC: tuple[type[BaseSpec], ...] = tuple([_process_technique_class(c) for c in ALL_TECHNIQUES])

if TYPE_CHECKING:
    TechniqueMethod = BaseSpec
else:
    TechniqueMethod = Union[tuple(ALL_TECHNIQUES_SPEC)]
    """A union of technique classes with an added type attribute."""


@dataclass
class TechniqueItem:
    """Configuration for applying a single anonymization technique.

    Attributes:
        method: Parameters defining the technique's configuration.
        columns: Input data columns to which the technique will be applied.
            If set to None, the technique is applied to all columns.
            An empty list is not allowed and will raise an error.
    """

    method: TechniqueMethod
    columns: list[str] | None

    # Utility mapping from a type to its corresponding spec class.
    _techniques_mapping: ClassVar[dict[TechniqueType, type[BaseSpec]]] = {
        _get_type_from_class(cls): cls for cls in ALL_TECHNIQUES_SPEC
    }

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "TechniqueItem":
        """Creates an instance of the class from a dictionary.

        Args:
            value: A dictionary where keys represent the attributes of the class
                and values are their corresponding values.

        Returns:
            An instance of the class populated with the data from the dictionary.
        """
        method_data: dict[str, Any] = value.get("method", {})
        if not method_data or "type" not in method_data:
            raise ValueError("Invalid input: 'method' field must have a 'type' key")

        method_type: TechniqueType = method_data.get("type", "")
        method_class: type[BaseSpec] | None = cls._techniques_mapping.get(method_type, None)
        if method_class is None:
            raise ValueError(f"Invalid input: unknown technique type '{method_type}'")

        method_kwargs: dict[str, Any] = method_data.copy()
        method_kwargs.pop("type")
        method: TechniqueMethod = method_class(**method_kwargs)
        columns: list[str] | None = value.get("columns", None)
        if columns == []:
            raise ValueError("Invalid input: the columns list cannot be empty.")
        return cls(method=method, columns=columns)

    def to_dict(self) -> dict[str, Any]:
        """Converts the instance of the class into a dictionary.

        Returns:
            dict: A dictionary where keys are attribute names and values
                are the corresponding attribute values of the object.
        """
        method_fields: dict[str, Any] = {
            name: getattr(self.method, name) for name in getattr(self.method, "_spec_fields")
        }
        method_fields["type"] = method_fields["type"].value
        return {
            "method": method_fields,
            "columns": self.columns,
        }


@dataclass
class Config:
    """Configuration for the high-level interface `aindo.anonymize.AnonymizationPipeline`.

    Attributes:
        steps:  A list of anonymization steps to be applied.
    """

    steps: list[TechniqueItem]

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "Config":
        """Creates an instance of the class from a dictionary.

        Args:
            value: A dictionary where keys represent the attributes of the class
                and values are their corresponding values.

        Returns:
            An instance of the class populated with the data from the dictionary.
        """
        steps_data = value.get("steps")
        if steps_data is None:
            raise ValueError("Invalid input: 'steps' not found")
        steps = [TechniqueItem.from_dict(item_data) for item_data in steps_data]
        return cls(steps=steps)

    def to_dict(self) -> dict[str, Any]:
        """Converts the instance of the class into a dictionary.

        Returns:
            dict: A dictionary where keys are attribute names and values
                are the corresponding attribute values of the object.
        """
        return {"steps": [s.to_dict() for s in self.steps]}



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Base class implementations for anonymization techniques."""

from abc import ABC, abstractmethod
from typing import ClassVar

import pandas as pd


class BaseTechnique(ABC):
    """Abstract base class for all anonymization techniques."""

    preserve_type: ClassVar[bool] = True

    @abstractmethod
    def anonymize(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """Applies the anonymization technique to the given data.

        Args:
            dataframe: The input data to be anonymized.

        Returns:
            The anonymized version of the input data.
        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        fields: str = ",".join([f"{k!s}={v!r}" for k, v in self.__dict__.items() if not k.startswith("_")])
        return f"{self.__class__.__name__}({fields})"


class BaseSingleColumnTechnique(BaseTechnique, ABC):
    """Abstract base class for anonymization techniques applied to a single column.

    Subclasses should implement the `anonymize_column` method,
    which defines the logic for anonymizing a single column.
    """

    @abstractmethod
    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        raise NotImplementedError()

    def anonymize_column(self, col: pd.Series) -> pd.Series:
        """Applies the anonymization technique to a single column.

        Args:
            col: The input data to be anonymized.

        Returns:
            The anonymized version of the input data.
        """
        out: pd.Series = self._apply_to_col(col)
        if self.preserve_type:
            return out.astype(col.dtype, copy=False)
        else:
            return out

    def anonymize(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """Applies the anonymization technique to a single-column dataframe.

        This is analogous to calling `anonymize_column()` on a single Pandas Series.
        It is a convenience method shared across all types of anonymizers.

        Args:
            dataframe: The input data. Must have exactly one column.

        Returns:
            The anonymized version of the input data.
        """
        if dataframe.shape[1] != 1:
            raise ValueError("Dataframe should have exactly one column")
        return pd.DataFrame(self.anonymize_column(dataframe.iloc[:, 0]), columns=dataframe.columns)



# SPDX-FileCopyrightText: 2025 Aindo SpA
#
# SPDX-License-Identifier: MIT

"""Implementation of the character masking technique."""

from typing import AnyStr, ClassVar, Generic, Literal

import pandas as pd

from aindo.anonymize.techniques.base import BaseSingleColumnTechnique

StartingDirection = Literal["left", "right"]


class CharacterMasking(BaseSingleColumnTechnique, Generic[AnyStr]):
    """Implements character masking.

    Character masking involves replacing, usually partially, the characters of a data value
    with a constant symbol.
    Full masking is achieved by setting `mask_length=-1`.

    Attributes:
        starting_direction: The direction in which masking starts. Default is "left".
        mask_length: The number of characters to mask.
            Set to -1 to mask the entire value.
            Defaults to 1.
        symbol: The symbol used for masking. Defaults to "*".
    """

    preserve_type: ClassVar[bool] = False

    mask_length: int = 1
    symbol: AnyStr = "*"  # type: ignore
    starting_direction: StartingDirection = "left"

    def __init__(
        self,
        mask_length: int = 1,
        symbol: AnyStr = "*",
        starting_direction: StartingDirection = "left",
    ) -> None:
        super().__init__()
        if (mask_length <= 0) and (mask_length != -1):
            raise ValueError("mask_length should be greater than zero or equal to -1")
        if len(symbol) != 1:
            raise ValueError("symbol should be a single character")

        self.starting_direction = starting_direction
        self.mask_length = mask_length
        self.symbol = symbol

    def _compute_mask(self, value: AnyStr) -> AnyStr:
        if self.mask_length == -1:
            return self.symbol * len(value)

        _mask_len: int = min(self.mask_length, len(value))
        mask: AnyStr = self.symbol * _mask_len
        if self.starting_direction == "right":
            return value[: -self.mask_length] + mask
        return mask + value[self.mask_length :]

    def _apply_to_col(self, col: pd.Series) -> pd.Series:
        return pd.Series([self._compute_mask(value) if pd.notna(value) else value for value in col], dtype="str")
